/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.xmi.util.xmidocumenter.ui;

import com.topcoder.xmi.util.xmidocumenter.modelelement.AttributeElement;
import com.topcoder.xmi.util.xmidocumenter.modelelement.ClassElement;
import com.topcoder.xmi.util.xmidocumenter.modelelement.ExceptionElement;
import com.topcoder.xmi.util.xmidocumenter.modelelement.OperationElement;
import com.topcoder.xmi.util.xmidocumenter.modelelement.ParameterElement;
import com.topcoder.xmi.util.xmidocumenter.modelelement.ReturnValueElement;
import com.topcoder.xmi.util.xmidocumenter.XMIDocumentFactory;
import com.topcoder.xmi.util.xmidocumenter.XMIDocument;
import com.topcoder.xmi.util.xmidocumenter.ModelElement;
import com.topcoder.xmi.util.xmidocumenter.XMIDocumenterException;
import com.topcoder.xmi.util.xmidocumenter.OutputFailedException;
import com.topcoder.xmi.util.xmidocumenter.ModificationFailedException;

import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.JTextArea;
import javax.swing.JToggleButton;
import javax.swing.JToolBar;
import javax.swing.JTree;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.filechooser.FileFilter;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeCellRenderer;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreeNode;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;


/**
 * This class provides a simple UI that uses XMI Documenter component to parse and edit an XMI document.
 *
 * @author traugust
 * @version 1.0
 */
public class XMIDocumenterUIDemo {
    /**
     * This is the window with in pixels.
     */
    private static final int WIDTH = 800;
    /**
     * This is the height of the window in pixels.
     */
    private static final int HEIGHT = 600;
    /**
     * This is the XMIDocumentFactory to be used for parsing input files.
     */
    private final XMIDocumentFactory factory;
    /**
     * This field stores the currently loaded XMIDocument, if any, <tt>null</tt> otherwise.
     */
    private XMIDocument currentDocument = null;
    /**
     * This is the main window of the application.
     */
    private final JFrame mainFrame;
    /**
     * This is the tree view of the application.
     */
    private final JTree jTree;
    /**
     * This is the model instance used by the tree view.
     */
    private DefaultTreeModel treeModel;
    /**
     * This is the editing area used to display and edit the documentation text.
     */
    private final JTextArea editingArea;
    /**
     * This is the save button, enabled when an XMIDocument has been loaded.
     */
    private final JButton saveButton;

    /**
     * This field contains the currently edited model element or <tt>null</tt> if no editing is in progress.
     */
    private ModelElement currentlyEditedModelElement;
    /**
     * This field contains the tee node that represents the currently edited model element.
     */
    private ModelElementTreeNode currentlyEditedTreeNode;
    /**
     * this mode flag triggers whether the renderer renders node that have missing documentation (and nodes that contain
     * nodes with missing documentation) to be marked red or not.
     */
    private boolean markMissingMode = false;

    /**
     * Creates a new XMIDocumenterUIDemo.
     */
    private XMIDocumenterUIDemo() {
        factory = new XMIDocumentFactory();

        mainFrame = new JFrame("XMI Documenter UI demo");
        final JToolBar toolbar = new JToolBar(JToolBar.HORIZONTAL);
        final JButton loadButton = new JButton("Load");
        toolbar.add(loadButton);
        loadButton.addActionListener(new ActionListener() {
            public void actionPerformed(final ActionEvent e) {
                load();
            }

        });
        saveButton = new JButton("Save");
        toolbar.add(saveButton);
        saveButton.addActionListener(new ActionListener() {
            public void actionPerformed(final ActionEvent e) {
                save();
            }
        });
        saveButton.setEnabled(false);

        final JToggleButton toggle = new JToggleButton("Mark undocumented nodes red");
        toolbar.add(toggle);
        toggle.addActionListener(new ActionListener() {
            public void actionPerformed(final ActionEvent e) {
                toggleMode(toggle.isSelected());
            }
        });

        final JSplitPane splitPane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT);

        editingArea = new JTextArea();
        editingArea.setEnabled(false);

        treeModel = new DefaultTreeModel(new DefaultMutableTreeNode("Nothing loaded"));
        jTree = new JTree(treeModel);
        jTree.setCellRenderer(new XMIDocumenterTreeRenderer());
        jTree.addTreeSelectionListener(new TreeSelectionListener() {
            public void valueChanged(final TreeSelectionEvent e) {
                final Object lastSelectedPathComponent = jTree.getLastSelectedPathComponent();
                endElementEditing();
                if (lastSelectedPathComponent instanceof ModelElementTreeNode) {
                    final ModelElementTreeNode modelElementTreeNode = (ModelElementTreeNode) lastSelectedPathComponent;
                    final ModelElement modelElement = modelElementTreeNode.getModelElement();
                    editModelElement(modelElement, modelElementTreeNode);
                }
            }
        });

        final Container contentPane = mainFrame.getContentPane();
        contentPane.setLayout(new BorderLayout());
        contentPane.add(toolbar, BorderLayout.NORTH);
        contentPane.add(splitPane, BorderLayout.CENTER);
        splitPane.setLeftComponent(new JScrollPane(jTree));
        splitPane.setRightComponent(new JScrollPane(editingArea));
    }

    /**
     * This is the main method that starts the application, creates a new XMIDocumenterUIDemo and shows it.
     *
     * @param args the args given on command line, ignored
     */
    public static void main(final String[] args) {
        new XMIDocumenterUIDemo().show();
    }

    /**
     * This is the action invoked by the load button, shows a file open dialog and loads the selected file.
     */
    private void load() {
        endElementEditing();
        final JFileChooser jFileChooser = new JFileChooser(new File("."));
        jFileChooser.setFileFilter(new XMIFileFilter());
        final int result = jFileChooser.showOpenDialog(mainFrame);
        if (result == JFileChooser.APPROVE_OPTION) {
            final File selectedFile = jFileChooser.getSelectedFile();
            try {
                final FileInputStream xmiStream = new FileInputStream(selectedFile);
                try {
                    currentDocument = factory.createDocument(xmiStream);
                    treeModel.setRoot(new XMIDocumentNode(currentDocument));
                    treeModel.reload();
                    saveButton.setEnabled(true);
                } finally {
                    xmiStream.close();
                }
            } catch (IOException e) {
                showErrorMessage(e, "An error occurred while loading the file");
            } catch (XMIDocumenterException e) {
                showErrorMessage(e, "An error occurred while loading the file");
            }
        }
    }

    /**
     * This is the action invoked by the save button, shows a save dialog and saves the currently edited XMIDocument to
     * the file selected.
     */
    private void save() {
        endElementEditing();
        final JFileChooser jFileChooser = new JFileChooser(new File("."));
        final int result = jFileChooser.showSaveDialog(mainFrame);
        if (result == JFileChooser.APPROVE_OPTION) {
            final File selectedFile = jFileChooser.getSelectedFile();
            try {
                final FileOutputStream out = new FileOutputStream(selectedFile);
                try {
                    currentDocument.writeTo(out);
                    treeModel.setRoot(new XMIDocumentNode(currentDocument));
                    treeModel.reload();
                } finally {
                    out.close();
                }
            } catch (IOException e) {
                showErrorMessage(e, "An error occurred while saving the file");
            } catch (OutputFailedException e) {
                showErrorMessage(e, "An error occurred while saving the file");
            }
        }
    }

    /**
     * This is the action of the toggle button for marking undocumented items in the tree, switches the mode flag and
     * repaints the tree.
     *
     * @param enabled the new mode to be set, <tt>true</tt> to mark undocumented items red, <tt>false</tt> to not do so
     */
    private void toggleMode(final boolean enabled) {
        markMissingMode = enabled;
        jTree.repaint();
    }

    /**
     * This method shows an error dialog displaying the given exception and using the given title, additionally printing
     * the stack trace of the given exception to the console.
     *
     * @param e     the exception to be displayed
     * @param title the title for the error dialog
     */
    private void showErrorMessage(final Exception e, final String title) {
        e.printStackTrace();
        JOptionPane.showMessageDialog(
            mainFrame, e.getMessage(),
            title,
            JOptionPane.ERROR_MESSAGE);
    }

    /**
     * This method is invoked upon selection of a ModelElementTreeNode. It ends editing any previously edited model
     * element and starts editing the newly selected ModelElement.
     *
     * @param modelElement         the modelElement represented by the given TreeNode
     * @param modelElementTreeNode the TreeNode selected
     */
    private void editModelElement(final ModelElement modelElement,
                                  final ModelElementTreeNode modelElementTreeNode) {
        endElementEditing();
        currentlyEditedTreeNode = modelElementTreeNode;
        currentlyEditedModelElement = modelElement;
        editingArea.setText(currentlyEditedModelElement.getDocumentationText());
        editingArea.setEnabled(true);
    }

    /**
     * This method ends any editing that might currently be active.
     */
    private void endElementEditing() {
        try {
            if (currentlyEditedTreeNode != null) {
                currentlyEditedTreeNode.invalidate();
                jTree.repaint();
            }
            if (currentlyEditedModelElement != null) {
                final String text = editingArea.getText();
                if (!text.equals(currentlyEditedModelElement.getDocumentationText())) {
                    currentlyEditedModelElement.setDocumentationText(text);
                }
                currentlyEditedModelElement = null;
                editingArea.setText("");
                editingArea.setEnabled(false);
            }
        } catch (ModificationFailedException e) {
            showErrorMessage(e, "An error occurred while editing the ModelElement");
        }
    }

    /**
     * This method shows the XMIDocumenterUIDemo instance application  window.
     */
    private void show() {
        mainFrame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        mainFrame.setSize(WIDTH, HEIGHT);
        mainFrame.setVisible(true);
    }

    /**
     * This is the abstract base tree node class for the tree used in the application.
     */
    private abstract static class BaseXMIDocumenterTreeNode implements TreeNode {
        /**
         * This list contains the child TreeNode instances.
         */
        private final List children = new ArrayList();
        /**
         * This field holds the parent tree node instance.
         */
        private final TreeNode parent;
        /**
         * This field caches the hasMissingDocumentation state of this TreeNode, <tt>null</tt> if no cached state
         * exist.
         */
        private Boolean cachedHasMissingDocumentation;

        /**
         * Creates a new BaseXMIDocumenterTreeNode with the given parent node.
         *
         * @param parent the parent node of this instance
         */
        protected BaseXMIDocumenterTreeNode(final TreeNode parent) {
            this.parent = parent;
        }

        /**
         * This method determines whether a node allows children. As the tree model is completely static after having
         * been built, this method returns <tt>true</tt> if the node has children and <tt>false</tt> otherwise.
         *
         * @return <tt>true</tt> when the node may have children, <tt>false</tt> otherwise
         */
        public boolean getAllowsChildren() {
            return !children.isEmpty();
        }

        /**
         * This method determines whether this node is a leaf.
         *
         * @return <tt>true</tt> if the node has no children, <tt>false</tt> if it has children
         */
        public boolean isLeaf() {
            return children.isEmpty();
        }

        /**
         * This method returns an enumeration of all child nodes of this node.
         *
         * @return an {@link java.util.Enumeration} of {@link javax.swing.tree.TreeNode}s that are the children of this
         *         node.
         */
        public Enumeration children() {
            return Collections.enumeration(children);
        }

        /**
         * This method returns the parent node of this node.
         *
         * @return the parent node of this node
         */
        public TreeNode getParent() {
            return parent;
        }

        /**
         * This method retrieves a child note at the index specified.
         *
         * @param childIndex the index of the child node to be retrieved
         *
         * @return the child node retrieved
         */
        public TreeNode getChildAt(final int childIndex) {
            return (TreeNode) children.get(childIndex);
        }

        /**
         * This method returns the index of the given child node in this node's children list.
         *
         * @param node the node to retrieve the index of
         *
         * @return the index of the given node, <tt>-1</tt> if given node is no child node of this node
         */
        public int getIndex(final TreeNode node) {
            return children.indexOf(node);
        }

        /**
         * This method returns the number of children this node has.
         *
         * @return the number of children of this node
         */
        public int getChildCount() {
            return children.size();
        }

        /**
         * This method returns a possibly cached state that expresses whether this node or any of its child nodes
         * represents a ModelElement with missing documentation.
         *
         * @return <tt>true</tt> if this node or any of its child nodes have missing documentation, <tt>false</tt>
         *         otherwise
         */
        public boolean getCachedMissingDocumentation() {
            if (cachedHasMissingDocumentation == null) {
                cachedHasMissingDocumentation = Boolean.valueOf(hasMissingDocumentation());
            }
            return cachedHasMissingDocumentation.booleanValue();
        }

        /**
         * This method invalidates the cached state of this node.
         */
        public void invalidate() {
            cachedHasMissingDocumentation = null;
            if (parent instanceof BaseXMIDocumenterTreeNode) {
                ((BaseXMIDocumenterTreeNode) parent).invalidate();
            }
        }

        /**
         * This method determines whether this node or any of its children represent a ModelElement with missing
         * documentation.
         *
         * @return <tt>true</tt> if this node or any of its child nodes have missing documentation, <tt>false</tt>
         *         otherwise
         */
        protected abstract boolean hasMissingDocumentation();

        /**
         * This method retrieves the list of children of this node.
         *
         * @return al List of TreeNodes that are the children of this node
         */
        protected List getChildren() {
            return children;
        }
    }

    /**
     * This class is a TreeNode that represents a ModelElement.
     */
    private static class ModelElementTreeNode extends BaseXMIDocumenterTreeNode {
        /**
         * This is the ModelElement represented by this node.
         */
        private final ModelElement element;

        /**
         * Creates a new ModelElementTreeNode with the given ModelElement and parent.
         *
         * @param element the ModelElement to be represented by this TreeNode
         * @param parent  the parent node of this node
         */
        ModelElementTreeNode(final ModelElement element, final TreeNode parent) {
            super(parent);
            this.element = element;
            final ModelElement[] ownedElements = element.getOwnedElements();
            for (int i = 0; i < ownedElements.length; i++) {
                final ModelElement ownedElement = ownedElements[i];
                getChildren().add(new ModelElementTreeNode(ownedElement, this));
            }
        }

        /**
         * This method determines whether this node or any of its children represent a ModelElement with missing
         * documentation.
         *
         * @return <tt>true</tt> if this node or any of its child nodes have missing documentation, <tt>false</tt>
         *         otherwise
         */
        public boolean hasMissingDocumentation() {
            final String documentationText = element.getDocumentationText();
            if (documentationText == null || documentationText.trim().length() == 0) {
                return true;
            }
            for (Iterator iterator = getChildren().iterator(); iterator.hasNext();) {
                final Object child = iterator.next();
                if (child instanceof BaseXMIDocumenterTreeNode) {
                    final BaseXMIDocumenterTreeNode modelElementTreeNode = (BaseXMIDocumenterTreeNode) child;
                    if (modelElementTreeNode.getCachedMissingDocumentation()) {
                        return true;
                    }
                }
            }
            return false;
        }

        /**
         * This method returns the ModelElement represented by this node.
         *
         * @return the ModelElement represented by this node
         */
        public ModelElement getModelElement() {
            return element;
        }
    }

    /**
     * An XMIDocumentNode is a TreeNode that represents an XMIDocument instance.
     */
    private static class XMIDocumentNode extends BaseXMIDocumenterTreeNode {
        /**
         * This map contains the PackageNode instances already created, maps from String (package name) to PackageNode.
         */
        private Map createdPackageNodes;

        /**
         * Creates a new XMIDocumentNode that represents the given XMIDocument.
         *
         * @param document the XMIDocument of this node
         */
        private XMIDocumentNode(final XMIDocument document) {
            super(null);
            createdPackageNodes = new HashMap();
            final ClassElement[] ownedElements = document.getClasses();
            for (int i = 0; i < ownedElements.length; i++) {
                final ClassElement classElement = ownedElements[i];
                final String packageName = classElement.getPackageName();
                final PackageNode packageNode = findPackageNode(packageName);
                packageNode.addChildNode(new ModelElementTreeNode(classElement, packageNode));
            }
        }

        /**
         * This utility method is used during constructor invocation for creation of new package nodes. When {@link
         * #createdPackageNodes} already contains a node representing the package, the node is returned, otherwise a new
         * node is created.
         *
         * @param packageName the package name for which to create a PackageNode
         *
         * @return the packageNode representing the given packageName
         */
        private PackageNode findPackageNode(final String packageName) {
            PackageNode packageNode = (PackageNode) createdPackageNodes.get(packageName);
            if (packageNode == null) {
                if (packageName.indexOf('.') == -1) {
                    packageNode = new PackageNode(packageName, this);
                    getChildren().add(packageNode);
                } else {
                    final String parentPackage = packageName.substring(0, packageName.lastIndexOf('.'));
                    final PackageNode parentPackageNode = findPackageNode(parentPackage);
                    packageNode = new PackageNode(packageName, parentPackageNode);
                    parentPackageNode.addChildNode(packageNode);
                }
                createdPackageNodes.put(packageName, packageNode);
            }
            return packageNode;
        }

        /**
         * This method determines whether this node or any of its children represent a ModelElement with missing
         * documentation.
         *
         * @return <tt>true</tt> if this node or any of its child nodes have missing documentation, <tt>false</tt>
         *         otherwise
         */
        public boolean hasMissingDocumentation() {
            for (Iterator iterator = getChildren().iterator(); iterator.hasNext();) {
                final PackageNode packageNode = (PackageNode) iterator.next();
                if (packageNode.getCachedMissingDocumentation()) {
                    return true;
                }
            }
            return false;
        }
    }

    /**
     * A PackageNode represents a package and can contain package nodes and ModelElementTreeNodes.
     */
    private static class PackageNode extends BaseXMIDocumenterTreeNode {
        /**
         * This is the name of the package represented by this instance.
         */
        private final String name;

        /**
         * Creates a new PackageNode representing the given package name.
         *
         * @param name   the name of the package to be represented by this node
         * @param parent the parent node of this node
         */
        PackageNode(final String name, final TreeNode parent) {
            super(parent);
            this.name = name;
        }

        /**
         * This method adds a new child node to this node.
         *
         * @param childNode the child node to be added
         */
        public void addChildNode(final TreeNode childNode) {
            getChildren().add(childNode);
        }

        /**
         * This method retrieves the package name of this package node.
         *
         * @return the package name represented by this node
         */
        public String getName() {
            return name;
        }

        /**
         * This method determines whether this node or any of its children represent a ModelElement with missing
         * documentation.
         *
         * @return <tt>true</tt> if this node or any of its child nodes have missing documentation, <tt>false</tt>
         *         otherwise
         */
        protected boolean hasMissingDocumentation() {
            for (Iterator iterator = getChildren().iterator(); iterator.hasNext();) {
                final Object child = iterator.next();
                if (child instanceof BaseXMIDocumenterTreeNode) {
                    if (((BaseXMIDocumenterTreeNode) child).getCachedMissingDocumentation()) {
                        return true;
                    }
                }
            }
            return false;
        }
    }

    /**
     * This class is the tree renderer used by he application.
     */
    private class XMIDocumenterTreeRenderer extends DefaultTreeCellRenderer {
        /**
         * This method renders the given node.
         *
         * @param tree     the parent tree of the node to be rendered
         * @param value    the node to be rendered
         * @param sel      whether the node is currently selected
         * @param expanded whether the node is currently expanded or collapsed
         * @param leaf     whether the given node is a leaf node
         * @param row      the row index of the node inside the tree
         * @param hasFocus whether the given node has focus
         *
         * @return the Component that represents the visual state to display the given node in
         */
        public Component getTreeCellRendererComponent(final JTree tree, final Object value, final boolean sel,
                                                      final boolean expanded,
                                                      final boolean leaf,
                                                      final int row, final boolean hasFocus) {
            final JLabel ret = (JLabel) super.getTreeCellRendererComponent(tree, value, sel, expanded, leaf, row,
                hasFocus);

            if (value instanceof XMIDocumentNode) {
                final XMIDocumentNode xmiDocumentNode = (XMIDocumentNode) value;
                if (markMissingMode && xmiDocumentNode.hasMissingDocumentation()) {
                    ret.setForeground(Color.RED);
                }
                ret.setText("XMI Document");
            } else if (value instanceof ModelElementTreeNode) {
                final ModelElementTreeNode modelElementTreeNode = (ModelElementTreeNode) value;
                if (markMissingMode && modelElementTreeNode.hasMissingDocumentation()) {
                    ret.setForeground(Color.RED);
                }
                final ModelElement modelElement = modelElementTreeNode.getModelElement();
                if (modelElement instanceof ClassElement) {
                    ret.setText("CLASS:" + modelElement.getName());
                } else if (modelElement instanceof OperationElement) {
                    ret.setText("OPERATION:" + modelElement.getName());
                } else if (modelElement instanceof ParameterElement) {
                    ret.setText("@PARAM:" + modelElement.getName());
                } else if (modelElement instanceof ReturnValueElement) {
                    ret.setText("@RETURN:" + modelElement.getName());
                } else if (modelElement instanceof ExceptionElement) {
                    ret.setText("@THROWS:" + modelElement.getName());
                } else if (modelElement instanceof AttributeElement) {
                    ret.setText("FIELD:" + modelElement.getName());
                }
            } else if (value instanceof PackageNode) {
                final PackageNode packageNode = (PackageNode) value;
                if (markMissingMode && packageNode.hasMissingDocumentation()) {
                    ret.setForeground(Color.RED);
                }
                ret.setText("PACKAGE:" + packageNode.getName());
            }
            return ret;
        }
    }

    /**
     * This filter is used by the open dialog. It filters the file view to only display XMI files and directories.
     */
    private static class XMIFileFilter extends FileFilter {
        /**
         * This method determines whether a file should be displayed in the file chooser.
         *
         * @param f the file to determine acceptance for
         *
         * @return <tt>true</tt> if the file is to be displayed in the file chooser, <tt>false</tt> otherwise
         */
        public boolean accept(final File f) {
            return f.getName().endsWith(".xmi") || f.isDirectory();
        }

        /**
         * This method returns the display name of this filter.
         *
         * @return the display name of this filter
         */
        public String getDescription() {
            return "XMI files (*.xmi)";
        }
    }
}
